__author__ = "Hugo Inzirillo"

from dataclasses import dataclass
from enum import Enum
from typing import List, Union


class Scope(Enum):
    READ = "READ"
    WRITE = "WRITE"


@dataclass
class ClientCredentials:
    client_id: str
    client_secret: str
    scope: Union[List[Scope], List[str]]

    def __post_init__(self):
        self.scope = [item.value for item in self.scope if isinstance(item, Scope)]



__author__ = "Hugo Inzirillo"

import json
import logging

from authlib.integrations.requests_client import OAuth2Session
from requests import Response
from requests.exceptions import HTTPError

from napbots.api.authentication.auth import ClientCredentials
from napbots.api.data.serializable import Formatter


class BaseClient:
    _base_url = None

    def __init__(self):
        self._client_credentials = None
        pass

    @property
    def credentials(self) -> ClientCredentials:
        return self._client_credentials

    @credentials.setter
    def credentials(self, _credentials: ClientCredentials):
        if isinstance(_credentials, ClientCredentials):
            self._client_credentials = _credentials

    @property
    def session(self) -> OAuth2Session:
        _session = OAuth2Session(
            self.credentials.client_id,
            self.credentials.client_secret,
            scope=self.credentials.scope,
        )

        _session.fetch_token(self._get_url("/oauth/token"))
        return _session

    def _get_url(self, url: str):
        return self._base_url + url

    @property
    def headers(self):
        return {"Content-type": "application/json", "Accept": "application/json"}

    @staticmethod
    def _handle_response(response: Response) -> Response:
        if response.status_code == 200:
            print(response.json())
            return response.json()
        else:
            print(response.json())
            return response

    def _get(self, route, **kwargs):
        url = self._get_url(route)
        response = self.session.get(url, headers=self.headers, **kwargs)
        return self._handle_response(response)

    def _post(self, route, data=None, **kwargs) -> Response:
        url = self._get_url(route)
        dumps = format(data, Formatter.JSON)
        print(f"send : {dumps}")
        response = self.session.post(url, data=dumps, headers=self.headers, **kwargs)
        return self._handle_response(response)



__author__ = "Hugo Inzirillo"

import sys


from datetime import datetime, timezone


from napbots.api.authentication import Scope
from napbots.api.client.client import client
from napbots.api.data.model import (
    EligibleComponentList,
    EligibleComponent,
    Strategy,
    UnderlyingList,
    ProductCode,
    Underlying,
    ExposureType,
)


if __name__ == "__main__":
    client = client(
        client_id="hougo", client_secret="", scope=[Scope.READ, Scope.WRITE]
    )
    now = datetime.now(tz=timezone.utc)
    component_list = EligibleComponentList()
    component_list.append(
        EligibleComponent(
            product_code=ProductCode.ETH_USD, exposure_type=ExposureType.LO
        )
    )
    strategy = Strategy("ALPHA", component_list, cron="0 0 1 0/1 * *")
    client.get_strategy("STRAT_NAPBOTS_HOUGO_ALPHA")
    client.create_strategy(strategy)



__author__ = "Hugo Inzirillo"

from datetime import datetime
from enum import Enum
from napbots.api.data.serializable import Serializable


def raise_error(_class):
    raise TypeError(f"components must be an instance of {_class}")


class ProductCode(Enum):
    BTC_USD = "BTC-USD"
    ETH_USD = "ETH-USD"
    ADA_USD = "ADA-USD"
    BNB_USD = "BNB-USD"
    MATIC_USD = "MATIC-USD"
    EOS_USD = "EOS-USD"
    XRP_USD = "XRP-USD"
    LTC_USD = "LTC-USD"
    XLM_USD = "XLM-USD"
    ZEC_USD = "ZEC-USD"
    BCH_USD = "BCH-USD"
    LINK_USD = "LINK-USD"
    SOL_USD = "SOL-USD"
    XTZ_USD = "XTZ-USD"
    DOT_USD = "DOT-USD"
    UNI_USD = "UNI-USD"


class ComponentList(list):
    pass


class UnderlyingList(list):
    pass


class SignalList(list):
    pass


class EligibleComponentList(list):
    pass


class ExposureType(Enum):
    LO = "LONG_ONLY"
    AR = "ABSOLUTE_RETURN"


class Data:
    """
    this class implement the global representation for data
    """

    def __repr__(self):
        kwargs = [f"{key}={value!r}" for key, value in self.__dict__.items() if key[0] != "_" or key[:2] != "__"]
        return "{}({})".format(type(self).__name__, "".join(kwargs))

    @classmethod
    def from_json(cls, **kwargs):
        raise NotImplemented

    @property
    def json(self):
        raise NotImplementedError


class Strategy(Data, Serializable):

    def __init__(
            self,
            label: str,
            components: EligibleComponentList,
            cron: str,
            strategy_id: str = None,
    ):
        if not isinstance(components, EligibleComponentList):
            raise_error(EligibleComponentList)
        self.strategy_id = strategy_id
        self.components = components
        self.label = label
        self.cron = cron

    @classmethod
    def from_json(cls, **kwargs):
        eligible_components = kwargs.get("eligibleComponents")
        eligible_component_list = EligibleComponentList(
            map(lambda x: EligibleComponent.from_json(**x), eligible_components))
        strategy_id = kwargs.get("strategyID")
        label = kwargs.get("label")
        cron = kwargs.get("reweigthCron")
        return cls(label, eligible_component_list, cron, strategy_id)

    @property
    def json(self):
        return dict(
            strategyID=self.strategy_id,
            eligibleComponents=[component.json for component in self.components],
            label=self.label,
            rebalancingCron=self.cron
        )


class Signal(Data, Serializable):

    def __init__(
            self,
            strategy_id: str,
            ts: datetime,
            underlyings: UnderlyingList

    ):
        if not isinstance(underlyings, UnderlyingList):
            raise_error(ComponentList)
        self.underlyings = underlyings
        self.strategy_id = strategy_id
        self.ts = ts

    @classmethod
    def from_json(cls, **kwargs):
        components = kwargs.get("componentsList")
        underlying_list = UnderlyingList(map(lambda x: Underlying.from_json(**x), components))
        strategy_id = kwargs.get("strategyID")
        ts = kwargs.get("ts")
        return cls(strategy_id, datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S"), underlying_list)

    @property
    def json(self):
        return dict(
            underlyingList=[underlying.json for underlying in self.underlyings],
            strategyID=self.strategy_id,
            rebalancingTs=self.format(self.ts)

        )


class Underlying(Data):
    def __init__(
            self,
            product_code: str,
            weight: float
    ):
        self.product_code = product_code
        self.weight = weight

    @classmethod
    def from_json(cls, **kwargs):
        product_code = kwargs.get("productCode")
        weight = kwargs.get("weight")
        return cls(product_code, weight)

    @property
    def json(self):
        return dict(
            productCode=self.product_code,
            weight=self.weight
        )


class EligibleComponent(Data, Serializable):
    __dict__ = {}

    def __init__(
            self,
            product_code: ProductCode,
            exposure_type: ExposureType
    ):
        if not isinstance(product_code, ProductCode):
            raise_error(ProductCode)

        if not isinstance(exposure_type, ExposureType):
            raise_error(ExposureType)

        self.product_code = product_code.value
        self.exposure_type = exposure_type.value

    @classmethod
    def from_json(cls, **kwargs):
        exposure_type = kwargs.get("exposureType")
        product_code = kwargs.get("productCode")
        if exposure_type is not None:
            exposure_type = ExposureType(exposure_type)
        if product_code is not None:
            product_code = ProductCode(product_code)
        return cls(product_code, exposure_type)

    @property
    def json(self):
        return dict(
            productCode=self.product_code,
            exposureType=self.exposure_type
        )



__author__ = "Hugo Inzirillo"

import json
from abc import ABCMeta, abstractmethod
from datetime import datetime, date
from enum import Enum
from functools import singledispatchmethod


class _Encoder(Enum):
    JSON = "JSON"


class Formatter(metaclass=ABCMeta):
    JSON = "JSON"

    @abstractmethod
    def format(self):
        raise NotImplementedError


class _DateTimeObjectFormatter(Formatter):
    @singledispatchmethod
    def format(self, arg):
        pass

    @format.register
    def _(self, arg: date):
        return arg.strftime("%Y-%m-%d")

    @format.register
    def _(self, arg: datetime):
        return arg.strftime("%Y-%m-%dT%H:%M:%S")



    @format.register
    def _(self, arg: str):
        return self.format(datetime.strptime(arg, "%Y-%m-%dT%H:%M:%S"))


class Serializable(_DateTimeObjectFormatter, metaclass=ABCMeta):
    @property
    @abstractmethod
    def json(self):
        raise NotImplementedError

    def __iter__(self):
        iterable = dict((x, y) for x, y in self.__dict__.items() if x[:2] != '__')
        iterable.update(self.__dict__)
        for x, y in iterable.items():
            yield x, y

    def __format__(self, encoder):
        if isinstance(encoder, str):
            if _Encoder(encoder) == _Encoder.JSON:
                return json.dumps(self.json)
            else:
                raise NotImplementedError("Others encoders are not available")
        else:
            raise TypeError(f'encoder should be an instance of {type(_Encoder)}')



__author__ = "Hugo Inzirillo"

from datetime import datetime
from typing import Union, List

from flask import Response

from napbots.api.authentication.auth import ClientCredentials, Scope
from napbots.api.base.base import BaseClient
from napbots.api.data.model import SignalList, Signal, Strategy, UnderlyingList


class NapbotsApiClient(BaseClient):
    _base_url = "https://api.napbots.com"

    def __init__(self, credentials: ClientCredentials):
        super().__init__()
        self.credentials = credentials

    def create_strategy(
            self,
            strategy: Strategy
    ) -> Response:
        return self._post("/v1/strategy", data=strategy)

    def get_strategy(
            self,
            strategy_id
    ) -> Strategy:
        json = self._get("/v1/strategy/{strategyID}".format(strategyID=strategy_id))
        return Strategy.from_json(**json.get("data"))

    def get_all_strategy(
            self,
            client_id
    ) -> list:
        json = self._get("/v1/strategy/all/for-owner/{client_id}".format(client_id=client_id))
        return list(map(lambda x: Strategy.from_json(**x), json.get("data")))

    def send_signal(
            self,
            strategy_id: str,
            ts: datetime,
            underlyings: UnderlyingList
    ):
        signal = Signal(strategy_id, ts, underlyings)
        return self._post("/v1/reweight", data=signal)


    def get_signals(
            self,
            strategy_id: str,
            start_ts: datetime,
            end_ts: datetime
    ) -> SignalList:
        route = "/v1/reweight/for-strategy/{strategyID}/between/{startTs}/{endTs}".format(
            strategyID=strategy_id,
            startTs=start_ts.strftime("%Y-%m-%dT%H:%M:%S"),
            endTs=end_ts.strftime("%Y-%m-%dT%H:%M:%S")
        )
        json = self._get(route=route)
        return SignalList(map(lambda x: Signal.from_json(**x), json.get("data")))

    def get_last_signal_executed(self, strategy_id: str) -> Signal:
        route = "/v1/reweight/last/for-strategy/{strategyId}".format(
            strategyId=strategy_id
        )
        json = self._get(route=route)
        return Signal.from_json(**json.get("data"))

    def get_last_signal_pushed(self, strategy_id: str, ts: datetime) -> Signal:
        route = "/v1/reweight/last/for-strategy/{strategyId}/rebalancing-ts/{rebalancingTs}".format(
            strategyId=strategy_id, rebalancingTs=ts.strftime("%Y-%m-%dT%H:%M:%S")
        )
        json = self._get(route=route)
        return Signal.from_json(**json.get("data"))


def client(
        client_id: str,
        client_secret: str,
        scope: Union[List[str], List[Scope]]
) -> NapbotsApiClient:
    credentials = ClientCredentials(client_id, client_secret, scope)
    return NapbotsApiClient(credentials)
